package zap

import (
	"context"
	"fmt"
	"testing"

	"git.perx.ru/perxis/perxis-go/id"
	"git.perx.ru/perxis/perxis-go/logs"
	"git.perx.ru/perxis/perxis-go/pkg/items"
	logzap "git.perx.ru/perxis/perxis-go/zap"
	"github.com/stretchr/testify/require"
	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"
)

func TestEntryEncoder_EncodeEntry(t *testing.T) {
	tests := []struct {
		name  string
		input struct {
			entry  zapcore.Entry
			fields []zapcore.Field
		}
		want *logs.Entry
	}{
		{
			name: "simple",
			input: struct {
				entry  zapcore.Entry
				fields []zapcore.Field
			}{
				entry: zapcore.Entry{Level: zapcore.InfoLevel, Message: "создан элемент коллекции"},
				fields: []zapcore.Field{
					zap.String("key", "val"), // будет проигнорировано
					logzap.Category("create"),
					logzap.Component("Items.Service"),
					logzap.Event("Items.Create"),
					logzap.Object("/spaces/WPNN/envs/9VGP/cols/GxNv/items/W0fl"),
					logzap.Caller(context.TODO(), logzap.WithObject("/users/PHVz")),
					logzap.Attr("any"),
					logzap.Tags("tag1", "tag2", "tag3"),
				},
			},
			want: &logs.Entry{
				Level:     logs.Level(zapcore.InfoLevel),
				Message:   "создан элемент коллекции",
				Category:  "create",
				Component: "Items.Service",
				Event:     "Items.Create",
				ObjectID:  id.MustObjectId("/spaces/WPNN/envs/9VGP/cols/GxNv/items/W0fl"),
				CallerID:  id.MustObjectId("/users/PHVz"),
				Attr:      "any",
				Tags:      []string{"tag1", "tag2", "tag3"},
			},
		},
	}

	enc := NewEntryEncoder()

	for _, tc := range tests {
		t.Run(tc.name, func(t *testing.T) {
			got, _ := enc.EncodeEntry(tc.input.entry, tc.input.fields)
			got.ID = tc.want.ID               // игнорируем ID
			got.Timestamp = tc.want.Timestamp // игнорируем Timestamp
			require.Equal(t, tc.want, got)
		})
	}
}

func BenchmarkEntryEncoderSimple(b *testing.B) {
	fields := []zapcore.Field{
		logzap.Event(items.EventCreate),
		logzap.Object(items.NewItem("WPNN", "9VGP", "GxNv", "W0fl", nil, nil)),
		logzap.Caller(context.TODO(), logzap.WithObject("/system")),
		logzap.Tags("tag1", "tag2", "tag3"),
	}

	enc := NewEntryEncoderSlow()
	for i := 0; i < b.N; i++ {
		_, _ = enc.EncodeEntry(zapcore.Entry{Message: fmt.Sprintf("Msg: %d", i)}, fields)
	}
}

func BenchmarkEntryEncoderUnknownFields(b *testing.B) {
	fields := []zapcore.Field{
		logzap.Event(items.EventCreate),
		logzap.Object(items.NewItem("WPNN", "9VGP", "GxNv", "W0fl", nil, nil)),
		logzap.Caller(context.TODO(), logzap.WithObject("/system")),
		logzap.Tags("tag1", "tag2", "tag3"),
	}

	for i := 0; i < 1000; i++ {
		fields = append(fields, zap.String(fmt.Sprintf("Key: %d", i), fmt.Sprintf("Value: %d", i)))
	}

	enc := NewEntryEncoderSlow()
	for i := 0; i < b.N; i++ {
		_, _ = enc.EncodeEntry(zapcore.Entry{Message: fmt.Sprintf("Msg: %d", i)}, fields)
	}
}

func BenchmarkEntryEncoderV2Simple(b *testing.B) {
	fields := []zapcore.Field{
		logzap.Event(items.EventCreate),
		logzap.Object(items.NewItem("WPNN", "9VGP", "GxNv", "W0fl", nil, nil)),
		logzap.Caller(context.TODO(), logzap.WithObject("/system")),
		logzap.Tags("tag1", "tag2", "tag3"),
	}

	enc := NewEntryEncoder()
	for i := 0; i < b.N; i++ {
		_, _ = enc.EncodeEntry(zapcore.Entry{Message: fmt.Sprintf("Msg: %d", i)}, fields)
	}
}

func BenchmarkEntryEncoderV2UnknownFields(b *testing.B) {
	fields := []zapcore.Field{
		logzap.Event(items.EventCreate),
		logzap.Object(items.NewItem("WPNN", "9VGP", "GxNv", "W0fl", nil, nil)),
		logzap.Caller(context.TODO(), logzap.WithObject("/system")),
		logzap.Tags("tag1", "tag2", "tag3"),
	}

	for i := 0; i < 1000; i++ {
		fields = append(fields, zap.String(fmt.Sprintf("Key: %d", i), fmt.Sprintf("Value: %d", i)))
	}

	enc := NewEntryEncoder()
	for i := 0; i < b.N; i++ {
		_, _ = enc.EncodeEntry(zapcore.Entry{Message: fmt.Sprintf("Msg: %d", i)}, fields)
	}
}
