package middleware

import (
	"context"
	"fmt"

	"git.perx.ru/perxis/perxis-go/id"
	"git.perx.ru/perxis/perxis-go/pkg/members"
	logzap "git.perx.ru/perxis/perxis-go/zap"
	"go.uber.org/zap"
)

type loggingMiddleware struct {
	logger *zap.Logger
	next   members.Members
}

func LoggingMiddleware(logger *zap.Logger) Middleware {
	return func(next members.Members) members.Members {
		return &loggingMiddleware{
			next:   next,
			logger: logger.With(logzap.Component("Members")),
		}
	}
}

func (m *loggingMiddleware) Set(ctx context.Context, orgId, userId string, role members.Role) (err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
		logzap.Event(members.EventSet),
		logzap.Object(id.NewOrganizationId(orgId)),
	)

	err = m.next.Set(ctx, orgId, userId, role)
	if err != nil {
		logger.Error(fmt.Sprintf("Failed to assign user '%s' to role '%s'", userId, role), zap.Error(err), logzap.Channels(logzap.Userlog, logzap.Syslog))
		return
	}

	logger.Info(fmt.Sprintf("User '%s' assigned to role '%s'", userId, role), logzap.Channels(logzap.Userlog))

	return err
}

func (m *loggingMiddleware) Get(ctx context.Context, orgId, userId string) (role members.Role, err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
		logzap.Object(id.NewRoleId(orgId, userId)),
	)

	role, err = m.next.Get(ctx, orgId, userId)
	if err != nil {
		logger.Error("Failed to get", zap.Error(err))
		return
	}

	return role, err
}

func (m *loggingMiddleware) Remove(ctx context.Context, orgId, userId string) (err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
		logzap.Event(members.EventRemove),
		logzap.Object(id.NewOrganizationId(orgId)),
	)

	err = m.next.Remove(ctx, orgId, userId)
	if err != nil {
		logger.Error(fmt.Sprintf("Failed to remove user '%s'", userId), zap.Error(err), logzap.Channels(logzap.Userlog, logzap.Syslog))
		return
	}

	logger.Info(fmt.Sprintf("User '%s' removed", userId), logzap.Channels(logzap.Userlog))

	return err
}

func (m *loggingMiddleware) RemoveAll(ctx context.Context, orgId string) (err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
		logzap.Event(members.EventRemoveAll),
		logzap.Object(id.NewOrganizationId(orgId)),
	)

	err = m.next.RemoveAll(ctx, orgId)
	if err != nil {
		logger.Error("Failed to remove all members", zap.Error(err), logzap.Channels(logzap.Userlog, logzap.Syslog))
		return
	}

	logger.Info("All members removed", logzap.Channels(logzap.Userlog))

	return err
}

func (m *loggingMiddleware) ListMembers(ctx context.Context, orgId string) (members []*members.Member, err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
	)

	members, err = m.next.ListMembers(ctx, orgId)
	if err != nil {
		logger.Error("Failed to list members", zap.Error(err))
		return
	}

	return members, err
}

func (m *loggingMiddleware) ListOrganizations(ctx context.Context, userId string) (organizations []*members.Member, err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
	)

	organizations, err = m.next.ListOrganizations(ctx, userId)
	if err != nil {
		logger.Error("Failed to list organizations", zap.Error(err))
		return
	}

	return organizations, err
}
